# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from os.path import getsize
from mock import patch
from nose.plugins.attrib import attr
from tempfile import NamedTemporaryFile

import pandas as pd
from xlrd import open_workbook

from vaxrank.cli import main as run_shell_script

from .testing_helpers import data_path

cli_args_for_b16_seqdata = [
    "--vcf", data_path("b16.f10/b16.vcf"),
    "--bam", data_path("b16.f10/b16.combined.bam"),
    "--vaccine-peptide-length", "25",
    "--mhc-predictor", "random",
    "--mhc-alleles", "H2-Kb,H2-Db",
    "--padding-around-mutation", "5",
    "--include-mismatches-after-variant"
]

cli_args_for_b16_seqdata_real_predictor = [
    "--vcf", data_path("b16.f10/b16.vcf"),
    "--bam", data_path("b16.f10/b16.combined.bam"),
    "--vaccine-peptide-length", "25",
    "--mhc-predictor", "netmhcpan",
    "--mhc-alleles", "H2-Kb,H2-Db",
    "--mhc-epitope-lengths", "8",
    "--padding-around-mutation", "5",
    "--include-mismatches-after-variant"
]


def test_ascii_report():
    with NamedTemporaryFile(mode="r") as f:
        ascii_args = cli_args_for_b16_seqdata + ["--output-ascii-report", f.name]
        run_shell_script(ascii_args)
        contents = f.read()
        lines = contents.split("\n")
        assert len(lines) > 0


def test_ascii_report_real_netmhc_predictor():
    with NamedTemporaryFile(mode="r") as f:
        ascii_args = cli_args_for_b16_seqdata_real_predictor + [
            "--output-ascii-report", f.name]
        run_shell_script(ascii_args)
        contents = f.read()
        lines = contents.split("\n")
        assert len(lines) > 0
        no_variants_text = 'No variants'
        assert no_variants_text not in contents


def test_json_report():
    with NamedTemporaryFile(mode="r") as f:
        json_args = cli_args_for_b16_seqdata + ["--output-json-file", f.name]
        run_shell_script(json_args)
        contents = f.read()
        lines = contents.split("\n")
        assert len(lines) > 0


def test_csv_report():
    with NamedTemporaryFile(mode="r") as f:
        csv_args = cli_args_for_b16_seqdata + ["--output-csv", f.name]
        run_shell_script(csv_args)
        contents = f.read()
        lines = contents.split("\n")
        assert len(lines) > 0


def test_all_variant_csv_report():
    with NamedTemporaryFile(mode="r") as f:
        all_csv_args = cli_args_for_b16_seqdata + [
            "--output-passing-variants-csv", f.name, "--output-csv", f.name + "ignored"]
        run_shell_script(all_csv_args)
        contents = f.read()
        lines = contents.split("\n")
        assert len(lines) > 0
        # make sure it can be a valid dataframe
        f.seek(0)
        df = pd.read_csv(f)
        assert len(df) > 0


def test_xlsx_report():
    with NamedTemporaryFile(mode="r") as f:
        xlsx_args = cli_args_for_b16_seqdata + ["--output-xlsx-report", f.name]
        run_shell_script(xlsx_args)
        book = open_workbook(f.name)
        assert book.nsheets > 0


def test_html_report():
    with NamedTemporaryFile(mode="r") as f:
        html_args = cli_args_for_b16_seqdata + ["--output-html", f.name]
        run_shell_script(html_args)
        contents = f.read()
        lines = contents.split("\n")
        assert len(lines) > 0


@attr('skip')  # want the ability to skip this test on some machines
def test_pdf_report():
    with NamedTemporaryFile(mode="rb") as f:
        pdf_args = cli_args_for_b16_seqdata + ["--output-pdf-report", f.name]
        run_shell_script(pdf_args)
        assert getsize(f.name) > 0


@patch('vaxrank.core_logic.VaxrankCoreLogic.vaccine_peptides_for_variant')
def test_report_no_peptides(mock_vaccine_peptides_for_variant):
    # simulate case where we have no epitopes for any variant
    mock_vaccine_peptides_for_variant.return_value = []
    with NamedTemporaryFile(mode="r") as f:
        html_args = cli_args_for_b16_seqdata + ["--output-csv", f.name]
        # test that this doesn't crash and that the CSV output is empty
        run_shell_script(html_args)
        contents = f.read()
        assert contents == ''


if __name__ == "__main__":
    test_csv_report()
    test_html_report()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, print_function, division
from collections import OrderedDict
from os.path import join, dirname

import pandas as pd


_ENSEMBL_GENE_ID_COLUMN_NAME = 'Ensembl Gene ID'
_MUTATION_COLUMN_NAME = 'Mutation'

_IFNG_RESPONSE_COLUMN_NAME = 'interferon_gamma_response'
_CLASS_I_MHC_COLUMN_NAME = 'class1_mhc_presentation_pathway'
_DRIVER_GENE_COLUMN_NAME = 'cancer_driver_gene'
_DRIVER_VARIANT_COLUMN_NAME = 'cancer_driver_variant'

_CURRENT_DIR = dirname(__file__)
_DATA_DIR = join(_CURRENT_DIR, "data")


class GenePathwayCheck(object):
    """
    This class is meant for use with gene/variant list files from
    https://github.com/openvax/gene-lists. Other files can be used as well, but
    need to follow a similar column structure. Most logic is based on Ensembl
    gene IDs.

    Parameters
    ----------
    interferon_gamma_response_csv : str, optional
        Local path to interferon-gamma response CSV file.

    class1_mhc_presentation_pathway_csv : str, optional
        Local path to MHC class I presentation pathway CSV file.

    cancer_driver_genes_csv : str, optional
        Local path to cancer driver genes CSV file.

    cancer_driver_variants_csv : str, optional
        Local path to cancer driver variants CSV file.
    """
    def __init__(
            self,
            interferon_gamma_response_csv=None,
            class1_mhc_presentation_pathway_csv=None,
            cancer_driver_genes_csv=None,
            cancer_driver_variants_csv=None):

        self.interferon_gamma_response_gene_set = self._load_set_from_csv(
            csv_path=interferon_gamma_response_csv,
            default_filename="interferon-gamma-response.csv",
            description="Interferon gamma response pathway",
            column_names=[_ENSEMBL_GENE_ID_COLUMN_NAME])

        self.class1_mhc_presentation_pathway_gene_set = self._load_set_from_csv(
            csv_path=class1_mhc_presentation_pathway_csv,
            default_filename="class1-mhc-presentation-pathway.csv",
            description="Class I MHC presentation pathway",
            column_names=[_ENSEMBL_GENE_ID_COLUMN_NAME])

        self.cancer_driver_genes_set = self._load_set_from_csv(
            csv_path=cancer_driver_genes_csv,
            default_filename="cancer-driver-genes.csv",
            description="Cancer driver genes",
            column_names=[_ENSEMBL_GENE_ID_COLUMN_NAME])
        # set of gene ID, variant description pairs
        self.cancer_driver_variants_set = self._load_set_from_csv(
            csv_path=cancer_driver_variants_csv,
            default_filename="cancer-driver-variants.csv",
            description="Driver variants",
            column_names=[_ENSEMBL_GENE_ID_COLUMN_NAME, _MUTATION_COLUMN_NAME])

    @classmethod
    def _load_set_from_csv(cls, csv_path, default_filename, description, column_names):
        if not csv_path:
            csv_path = join(_DATA_DIR, default_filename)
        df = pd.read_csv(csv_path)
        columns = []
        for column_name in column_names:
            if column_name not in df.columns:
                raise ValueError("%s file (%s) needs column '%s'" % (
                    description,
                    csv_path,
                    column_name))
            columns.append(df[column_name].values)
        if len(columns) == 1:
            return set(columns[0])
        else:
            return set(zip(*columns))

    def make_variant_dict(self, variant):
        """
        Returns a dictionary of boolean values, depending on whether we see this
        variant in any relevant pathway or cancer driver files.

        Parameters
        ----------
        variant : varcode.Variant
            Variant object to evaluate
        """
        effect_description = variant.effects().top_priority_effect().short_description
        overlapping_gene_ids = variant.gene_ids
        variant_dict = OrderedDict()
        variant_dict[_IFNG_RESPONSE_COLUMN_NAME] = any([
            gene_id in self.interferon_gamma_response_gene_set
            for gene_id in overlapping_gene_ids
        ])
        variant_dict[_CLASS_I_MHC_COLUMN_NAME] = any([
            gene_id in self.class1_mhc_presentation_pathway_gene_set
            for gene_id in overlapping_gene_ids
        ])
        variant_dict[_DRIVER_GENE_COLUMN_NAME] = any([
            gene_id in self.cancer_driver_genes_set
            for gene_id in overlapping_gene_ids
        ])

        variant_dict[_DRIVER_VARIANT_COLUMN_NAME] = any([
            (gene_id, effect_description) in self.cancer_driver_variants_set
            for gene_id in overlapping_gene_ids
        ])
        return variant_dict



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, print_function, division
import os
import logging

import six
from datacache import get_data_dir
import shellinford


logger = logging.getLogger(__name__)


def fm_index_path(genome):
    """
    Returns a path for cached reference peptides, for the given genome.
    """
    # if $VAXRANK_REF_PEPTIDES_DIR is set, that'll be the location of the cache
    cache_dir = get_data_dir(envkey='VAXRANK_REF_PEPTIDES_DIR')
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)

    return os.path.join(cache_dir, '%s_%d_%d.fm' % (
        genome.species.latin_name, genome.release, 2 if six.PY2 else 3))


def generate_protein_sequences(genome):
    """
    Generator whose elements are protein sequences from the given genome.

    Parameters
    ----------
    genome : pyensembl.EnsemblRelease
        Input genome to load for reference peptides
    """
    for t in genome.transcripts():
        if t.is_protein_coding:
            protein_sequence = t.protein_sequence
            if six.PY2:
                # shellinford on PY2 seems to sometimes fail with
                # unicode strings
                protein_sequence = protein_sequence.encode("ascii")
            yield protein_sequence


def load_reference_peptides_index(genome, force_reload=False):
    """
    Loads the FM index containing reference peptides.

    Parameters
    ----------
    genome : pyensembl.EnsemblRelease
        Input genome to load for reference peptides

    force_reload : bool, optional
        If true, will recompute index for this genome even if it already exists.

    Returns
    -------
    fm : shellinford.FMIndex
        Index populated with reference peptides from the genome
    """
    path = fm_index_path(genome)
    if force_reload or not os.path.exists(path):
        logger.info("Building FM index at %s", path)
        fm = shellinford.FMIndex()
        fm.build(generate_protein_sequences(genome), path)
        logger.info("Done building FM index")
        return fm
    return shellinford.FMIndex(filename=path)


class ReferenceProteome(object):
    def __init__(self, genome):
        self.fm_index = load_reference_peptides_index(genome)

    def contains(self, kmer):
        return len(self.fm_index.search(kmer)) > 0



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, print_function, division
from collections import OrderedDict
import traceback
import logging

import numpy as np
from serializable import Serializable

from .reference_proteome import ReferenceProteome


logger = logging.getLogger(__name__)


class EpitopePrediction(Serializable):
    def __init__(
            self,
            allele,
            peptide_sequence,
            wt_peptide_sequence,
            ic50,
            wt_ic50,
            percentile_rank,
            prediction_method_name,
            overlaps_mutation,
            source_sequence,
            offset,
            occurs_in_reference):
        self.allele = allele
        self.peptide_sequence = peptide_sequence
        self.wt_peptide_sequence = wt_peptide_sequence
        self.length = len(peptide_sequence)
        self.ic50 = ic50
        self.wt_ic50 = wt_ic50
        self.percentile_rank = percentile_rank
        self.prediction_method_name = prediction_method_name
        self.overlaps_mutation = overlaps_mutation
        self.source_sequence = source_sequence
        self.offset = offset
        self.overlaps_mutation = overlaps_mutation
        self.occurs_in_reference = occurs_in_reference

    @classmethod
    def from_dict(cls, d):
        """
        Deserialize EpitopePrediction from a dictionary of keywords.
        """
        d = d.copy()
        if "length" in d:
            # length argument removed in version 1.1.0
            del d["length"]
        return cls(**d)

    def logistic_epitope_score(
            self,
            midpoint=350.0,
            width=150.0,
            ic50_cutoff=5000.0):  # TODO: add these default values into CLI as arguments
        """
        Map from IC50 values to score where 1.0 = strong binder, 0.0 = weak binder
        Default midpoint and width for logistic determined by max likelihood fit
        for data from Alessandro Sette's 1994 paper:

           "The relationship between class I binding affinity
            and immunogenicity of potential cytotoxic T cell epitopes.

        TODO: Use a large dataset to find MHC binding range predicted to #
        correlate with immunogenicity
        """
        if self.ic50 >= ic50_cutoff:
            return 0.0

        rescaled = (float(self.ic50) - midpoint) / width
        # simplification of 1.0 - logistic(x) = logistic(-x)
        logistic = 1.0 / (1.0 + np.exp(rescaled))

        # since we're scoring IC50 values, let's normalize the output
        # so IC50 near 0.0 always returns a score of 1.0
        normalizer = 1.0 / (1.0 + np.exp(-midpoint / width))

        return logistic / normalizer

    def slice_source_sequence(self, start_offset, end_offset):
        """

        Parameters
        ----------
        start_offset : int

        end_offset : int

        Return EpitopePrediction object with source sequence and offset
        adjusted. If this slicing would shorten the mutant peptide, then
        return None.
        """
        if self.offset < start_offset:
            # this peptide starts before the requested slice through the
            # source sequence
            return None

        if self.offset + self.length > end_offset:
            # this peptide goes beyond the end of the requested slice
            # through the source sequence
            return None

        return EpitopePrediction(
            allele=self.allele,
            peptide_sequence=self.peptide_sequence,
            wt_peptide_sequence=self.wt_peptide_sequence,
            ic50=self.ic50,
            wt_ic50=self.wt_ic50,
            percentile_rank=self.percentile_rank,
            prediction_method_name=self.prediction_method_name,
            overlaps_mutation=self.overlaps_mutation,
            source_sequence=self.source_sequence[start_offset:end_offset],
            offset=self.offset - start_offset,
            occurs_in_reference=self.occurs_in_reference)


def slice_epitope_predictions(
        epitope_predictions,
        start_offset,
        end_offset):
    """
    Return subset of EpitopePrediction objects which overlap the given interval
    and slice through their source sequences and adjust their offset.
    """
    return [
        p.slice_source_sequence(start_offset, end_offset)
        for p in epitope_predictions
        if p.offset >= start_offset and p.offset + p.length <= end_offset
    ]

def predict_epitopes(
        mhc_predictor,
        protein_fragment,
        min_epitope_score=0.0,
        genome=None):
    """
    Parameters
    ----------
    mhc_predictor : mhctools.BasePredictor
        Object with predict_peptides method

    protein_fragment : MutantProteinFragment

    peptide_length : list of int
        Lengths of peptides to make pMHC binding predictions for

    min_epitope_score : float
        Ignore peptides with binding predictions whose normalized score is less
        than this.

    genome : pyensembl.Genome
        Genome whose proteome to use for reference peptide filtering

    Returns an OrderedDict of EpitopePrediction objects, keyed by a
    (peptide sequence, allele) tuple, that have a normalized score greater
    than min_epitope_score.

    Uses the input genome to evaluate whether the epitope occurs in reference.
    """
    results = OrderedDict()
    reference_proteome = ReferenceProteome(genome)

    # sometimes the predictors will fail, and we don't want to crash vaxrank in that situation
    # TODO: make more specific or remove when we fix error handling in mhctools
    try:
        mhctools_binding_predictions = mhc_predictor.predict_subsequences(
            {protein_fragment.gene_name: protein_fragment.amino_acids})
    except:
        logger.error(
            'MHC prediction errored for protein fragment %s, with traceback: %s',
            protein_fragment, traceback.format_exc())
        return results

    # compute the WT epitopes for each mutant fragment's epitopes; mutant -> WT
    wt_peptides = {}
    for binding_prediction in mhctools_binding_predictions:
        peptide = binding_prediction.peptide
        peptide_length = binding_prediction.length
        peptide_start_offset = binding_prediction.offset
        peptide_end_offset = peptide_start_offset + peptide_length

        overlaps_mutation = protein_fragment.interval_overlaps_mutation(
            start_offset=peptide_start_offset,
            end_offset=peptide_end_offset)

        if overlaps_mutation:
            full_reference_protein_sequence = (
                protein_fragment.predicted_effect().original_protein_sequence
            )
            global_epitope_start_pos = (
                protein_fragment.global_start_pos() + peptide_start_offset
            )
            wt_peptide = full_reference_protein_sequence[
                global_epitope_start_pos:global_epitope_start_pos + peptide_length]
            wt_peptides[peptide] = wt_peptide

    wt_predictions = []
    try:
        # filter to minimum peptide lengths
        valid_wt_peptides = [
            x for x in wt_peptides.values() if len(x) >= mhc_predictor.min_peptide_length
        ]
        if len(valid_wt_peptides) > 0:
            wt_predictions = mhc_predictor.predict_peptides(valid_wt_peptides)
    except:
        logger.error(
            'MHC prediction for WT peptides errored, with traceback: %s',
            traceback.format_exc())

    # break it out: (peptide, allele) -> prediction
    wt_predictions_grouped = {
        (wt_prediction.peptide, wt_prediction.allele): wt_prediction
        for wt_prediction in wt_predictions
    }

    # convert from mhctools.BindingPrediction objects to EpitopePrediction
    # which differs primarily by also having a boolean field
    # 'overlaps_mutation' that indicates whether the epitope overlaps
    # mutant amino acids or both sides of a deletion
    num_total = 0
    num_occurs_in_reference = 0
    num_low_scoring = 0
    for binding_prediction in mhctools_binding_predictions:
        num_total += 1
        peptide = binding_prediction.peptide
        peptide_length = binding_prediction.length
        peptide_start_offset = binding_prediction.offset
        peptide_end_offset = peptide_start_offset + peptide_length

        overlaps_mutation = protein_fragment.interval_overlaps_mutation(
            start_offset=peptide_start_offset,
            end_offset=peptide_end_offset)

        occurs_in_reference = reference_proteome.contains(peptide)
        if occurs_in_reference:
            logger.debug('Peptide %s occurs in reference', peptide)
            num_occurs_in_reference += 1

        # compute WT epitope sequence, if this epitope overlaps the mutation
        if overlaps_mutation:
            wt_peptide = wt_peptides[peptide]
            wt_prediction = wt_predictions_grouped.get(
                (wt_peptide, binding_prediction.allele))
            wt_ic50 = None
            if wt_prediction is None:
                # this can happen in a stop-loss variant: do we want to check that here?
                if len(wt_peptide) < mhc_predictor.min_peptide_length:
                    logger.info(
                        'No prediction for too-short WT epitope %s: possible stop-loss variant',
                        wt_peptide)
            else:
                wt_ic50 = wt_prediction.value

        else:
            wt_peptide = peptide
            wt_ic50 = binding_prediction.value

        epitope_prediction = EpitopePrediction(
            allele=binding_prediction.allele,
            peptide_sequence=peptide,
            wt_peptide_sequence=wt_peptide,
            ic50=binding_prediction.value,
            wt_ic50=wt_ic50,
            percentile_rank=binding_prediction.percentile_rank,
            prediction_method_name=binding_prediction.prediction_method_name,
            overlaps_mutation=overlaps_mutation,
            source_sequence=protein_fragment.amino_acids,
            offset=peptide_start_offset,
            occurs_in_reference=occurs_in_reference)

        if epitope_prediction.logistic_epitope_score() >= min_epitope_score:
            key = (epitope_prediction.peptide_sequence, epitope_prediction.allele)
            results[key] = epitope_prediction
        else:
            num_low_scoring += 1

    logger.info(
        "%d total peptides: %d occur in reference, %d failed score threshold",
        num_total,
        num_occurs_in_reference,
        num_low_scoring)
    return results



from serializable import Serializable

class PatientInfo(Serializable):
    def __init__(
            self,
            patient_id,
            vcf_paths,
            bam_path,
            mhc_alleles,
            num_somatic_variants,
            num_coding_effect_variants,
            num_variants_with_rna_support,
            num_variants_with_vaccine_peptides):
        self.patient_id = patient_id
        self.vcf_paths = vcf_paths
        self.bam_path = bam_path
        self.mhc_alleles = mhc_alleles
        self.num_somatic_variants = num_somatic_variants
        self.num_coding_effect_variants = num_coding_effect_variants
        self.num_variants_with_rna_support = num_variants_with_rna_support
        self.num_variants_with_vaccine_peptides = num_variants_with_vaccine_peptides
